# coding=utf-8
"""Utility functions for calculating Adaptive Thermal Comfort."""
from __future__ import division

import math
import sys
if (sys.version_info > (3, 0)):
    xrange = range


def adaptive_comfort_ashrae55(t_prevail, to):
    """Get adaptive comfort criteria according to ASHRAE-55.

    Note:
        [1] ASHRAE Standard 55 (2017). Thermal Environmental Conditions
        for Human Occupancy. Atlanta Georgoa: American Society of Heating,
        Refrigerating and Air Conditioning Engineers.

    Args:
        t_prevail: The prevailing outdoor temperature [C].  For the ASHRAE-55 adaptive
            comfort model, this is typically the average monthly outdoor temperature.
        to: Operative temperature [C]

    Returns:
        A dictionary containing results with the following keys

        -   to : Operative Temperature [C].
        -   t_comf : Adaptive comfort neutral temperature (desired by occupants) [C].
        -   deg_comf: The difference between the operative temperature (to)
            and the adaptive comfort neutral temperature (t_comf) [C].
            Negative values indicate cool conditions and positive values
            indicate warm conditions.
    """
    # fix upper and lower outdoor temperatures if outside the range of the model
    if t_prevail < 10.:
        t_prevail = 10.
    elif t_prevail > 33.5:
        t_prevail = 33.5

    # get the neutral temperature
    t_comf = neutral_temperature_ashrae55(t_prevail)
    result = {}
    result['to'] = to
    result['t_comf'] = t_comf
    result['deg_comf'] = to - t_comf

    return result


def adaptive_comfort_en15251(t_prevail, to):
    """Get adaptive comfort criteria according to EN-15251.

    Note:
        [1] CEN (2007) Standard EN15251. Indoor Environmental Input Parameters for
        Design and Assessment of Energy Performance of Buildings: Addressing indoor
        air quality, thermal environment, lighting and acoustics, Brussels: Comite
        Europeen de Normalisation.

    Args:
        t_prevail: The prevailing outdoor temperature [C].  For the EN-15251 adaptive
            comfort model, this is typically the exponentially weighted running mean
            of the outdoor temperature over the past week.  Use the weighted_running_mean
            functions to compute this value.
        to: Operative temperature [C]

    Returns:
        A dictionary containing results with the following keys

        -   to : Operative Temperature [C].
        -   t_comf : Adaptive comfort neutral temperature (desired by occupants) [C].
        -   deg_comf: The difference between the operative temperature (to)
            and the adaptive comfort neutral temperature (t_comf) [C].
            Negative values indicate cool conditions and positive values
            indicate warm conditions.
    """
    # fix upper and lower outdoor temperatures if outside the range of the model
    if t_prevail < 10.:
        t_prevail = 10.
    elif t_prevail > 30.:
        t_prevail = 30.

    # get the neutral temperature
    t_comf = neutral_temperature_en15251(t_prevail)
    result = {}
    result['to'] = to
    result['t_comf'] = t_comf
    result['deg_comf'] = to - t_comf

    return result


def adaptive_comfort_conditioned(t_prevail, to, conditioning, model):
    """Get adaptive comfort using SCATs neutral temp function for heated/cooled operation.

    Note that the use of adaptive comfort methods in conditioned buildings is not
    an official part of either ASHRAE-55 or EN-15251. For more information on how
    adaptive comfort methods can be applied to conditioned buildings, see the
    neutral_temperature_conditioned function.

    Args:
        t_prevail: The prevailing outdoor temperature [C].  For the EN-15251 adaptive
            comfort model, this is typically the exponentially weighted running mean
            of the outdoor temperature over the past week.  Use the weighted_running_mean
            functions to compute this value.
        to: Operative temperature [C]
        conditioning: A number between 0 and 1 that represents how "conditioned" vs.
            "free-running" the building is.

            * 0 = free-running (completely passive with no air conditioning)
            * 1 = conditioned (no operable windows and fully air conditioned)

        model: The comfort standard, which will be used to represent the "free-running"
            function.  Chose from: 'EN-15251', 'ASHRAE-55'.

    Returns:
        A dictionary containing results with the following keys

        -   to : Operative Temperature [C].
        -   t_comf : Adaptive comfort neutral temperature (desired by occupants) [C].
        -   deg_comf: The difference between the operative temperature (to)
            and the adaptive comfort neutral temperature (t_comf) [C].
            Negative values indicate cool conditions and positive values
            indicate warm conditions.
    """
    # fix upper and lower outdoor temperatures if outside the range of the model
    if t_prevail < 10.0:
        t_prevail = 10.0
    elif t_prevail > 30:
        t_prevail = 30

    # get the neutral temperature
    t_comf = neutral_temperature_conditioned(t_prevail, conditioning, model)
    result = {}
    result['to'] = to
    result['t_comf'] = t_comf
    result['deg_comf'] = to - t_comf

    return result


def adaptive_comfort_conditioned_function(conditioning, model):
    """Get an adaptive_comfort_conditioned function with pre-set conditioning/model.

    Args:
        conditioning: A number between 0 and 1 that represents how "conditioned" vs.
            "free-running" the building is.

            * 0 = free-running (completely passive with no air conditioning)
            * 1 = conditioned (no operable windows and fully air conditioned)

        model: The comfort standard, which will be used to represent the "free-running"
            function.  Chose from: 'EN-15251', 'ASHRAE-55'.
    """
    def comfort_funct(t_prevail, to):
        return adaptive_comfort_conditioned(t_prevail, to, conditioning, model)
    return comfort_funct


def t_operative(ta, tr):
    """Get operative temperature from air and radiant temperature.

    Args:
        ta: Air temperature [C]
        tr: Mean radiant temperature [C]

    Return:
        Operative temperature [C]
    """
    return (ta + tr) / 2


def neutral_temperature_ashrae55(t_prevail):
    """Get the neutral temperature (desired by occupants) according to ASHRAE-55.

    Note:
        [1] de Dear, R.J. and Brager, G.S. (2002) Thermal comfort in naturally
        ventilated buildings: Revisions to ASHRAE Standard 55.
        Energy and Buildings 34(6), 549-61.

        [2] de Dear, R.J. (1998) A global database of thermal comfort experiments.
        ASHRAE Technical data bulletin 14(1), 15-26.

    Args:
        t_prevail: The prevailing outdoor temperature [C].  For the ASHRAE-55 adaptive
            comfort model, this is typically the average monthly outdoor temperature.

    Return:
        The desired neutral temperature for the input previaling outdoor temperature.
    """
    return 0.31 * t_prevail + 17.8


def neutral_temperature_en15251(t_prevail):
    """Get the neutral temperature (desired by occupants) according to EN-15251.

    Note:
        [1] CIBSE (2006) Environmental Criteria for Design, Chapter 1: Environmental
        Design: CIBSE Guide A. London: Chartered Institution of Building
        Services Engineers.

        [2] Nicol, F. and McCartney, K. (2001) Final Report (Public) Smart Controls
        and Thermal Comfort (SCATs). Report to the European Comission of the Smart
        Controls and Thermal Comfort project. Oxford: Oxford Brokes University.

    Args:
        t_prevail: The prevailing outdoor temperature [C].  For the EN-15251 adaptive
            comfort model, this is typically the exponentially weighted running mean
            of the outdoor temperature over the past week.  Use the weighted_running_mean
            functions to compute this value.

    Return:
        The desired neutral temperature for the input previaling outdoor temperature.
    """
    return 0.33 * t_prevail + 18.8


def neutral_temperature_conditioned(t_prevail, conditioning, model='EN-15251'):
    """Get the neutral temperature for a conditioned or partly conditioned building.

    Note that the use of adaptive comfort methods in conditioned buildings is not an
    official part of any standard. Both the American ASHRAE-55 standard and the
    European EN-15251 standard state that the adaptive model should only be used
    when the following criteria are met:

    (a) There is no mechanical cooling or heating system in operation
        (in the case of ASHRAE-55, no mechanical cooling system is installed
        and no heating system is in operation)

    (b) Metabolic rates of occupants range from 1.0 to 1.3 met

    (c) Occupants are allowed to freely adapt their clothing insulation
        (in the case of ASHRAE-55, occupants must specifically be allowed
        to adapt clothing within a range at least as wide as 0.5 - 1.0 clo)

    However, the SCATs project[1], from which EN-15251 is derived, involved the survey
    of conditioned buildings and a neutral temperature function was obtained for
    heated and cooled modes of operation.  While the coefficient of determination (aka.
    R squared) of this function was not as strong as that for free-running buildings,
    it appears to be stronger than PMV calculated from the observed conditions in the
    SCATs data set[2].  Accordingly, it has been published in CIBSE guide[3] and has
    been included here along with methods to calculate netural temperature functions
    in between free-running and heated/cooled conditions.

    Note:
        [1] Nicol, F. and McCartney, K. (2001) Final Report (Public) Smart Controls
        and Thermal Comfort (SCATs). Report to the European Comission of the Smart
        Controls and Thermal Comfort project. Oxford: Oxford Brokes University.

        [2] Humphreys, M., Nicol, F. and Roaf, S. (2016) Adaptive Thermal Comfort:
        Foundations and Analysis. Routledge. Chapter 14: PMV and the results of
        field studies.

        [3] CIBSE (2006) Environmental Criteria for Design, Chapter 1: Environmental
        Design: CIBSE Guide A. London: Chartered Institution of Building
        Services Engineers.

    Args:
        t_prevail: The prevailing outdoor temperature [C].
        conditioning: A number between 0 and 1 that represents how "conditioned" vs.
            "free-running" the building is.

            * 0 = free-running (completely passive with no air conditioning)
            * 1 = conditioned (no operable windows and fully air conditioned)

        model: The comfort standard, which will be used to represent the "free-running"
            function.  Chose from: 'EN-15251', 'ASHRAE-55'.

    Return:
        The desired neutral temperature for the input previaling outdoor temperature.

    """
    if conditioning == 1:
        t_comf = 0.09 * t_prevail + 22.6
    elif model == 'ASHRAE-55':
        inv_conditioning = 1 - conditioning
        t_comf = ((0.09 * conditioning) + (0.31 * inv_conditioning)) * t_prevail + \
            ((22.6 * conditioning) + (17.8 * inv_conditioning))
    elif model == 'EN-15251':
        inv_conditioning = 1 - conditioning
        t_comf = ((0.09 * conditioning) + (0.33 * inv_conditioning)) * t_prevail + \
            ((22.6 * conditioning) + (18.8 * inv_conditioning))
    else:
        raise ValueError('Adaptive comfort model type {} not recognized. '
                         'Choose: EN-15251 or ASHRAE-55'.format(model))
    return t_comf


def cooling_effect_ashrae55(vel, to):
    """Get ASHRAE-55 cooling effect as a result of elevated air speed.

    Args:
        vel: Relative air velocity [m/s]
        to : Operative Temperature [C]

    Returns:
        ce -- Cooling effect as a result of elevated air speed [C]
    """
    ce = 0
    if vel >= 0.6 and to >= 25:
        if vel < 0.9:
            ce = 1.2
        elif vel < 1.2:
            ce = 1.8
        elif vel >= 1.2:
            ce = 2.2
    return ce


def cooling_effect_en15251(vel, to):
    """Get EN-15251 cooling effect as a result of elevated air speed.

    Args:
        vel: Relative air velocity [m/s]
        to : Operative Temperature [C]

    Returns:
        ce -- Cooling effect as a result of elevated air speed [C]
    """
    ce = 0
    if vel >= 0.2 and to >= 25:
        ce = 1.7856 * math.log(vel) + 2.9835
    return ce


def ashrae55_neutral_offset_from_ppd(ppd=90):
    """Get acceptable offset from neutral temperature given the ASHRAE-55 PPD limit.

    Args:
        ppd: The acceptable limit of Percentage of People Dissatisfied (PPD).
            Usually, this is 90% but it can be 80% in some cases.

    Returns:
        offset -- Acceptable temperature offset from neutral temperature [C]
    """
    assert 0 <= ppd <= 100, 'ppd must be between 0 and 100. Got {}'.format(ppd)
    return -0.1 * ppd + 11.5


def en15251_neutral_offset_from_comfort_class(comf_class):
    """Get acceptable offset from neutral temperature given the EN-15251 comfort class.

    Args:
        comf_class: An integer representing the EN-15251 comfort class.
            Choose from: 1, 2, 3 (the higher the class, the greater the offset)

    Returns:
        offset -- Acceptable temperature offset from neutral temperature [C]
    """
    if comf_class == 1:
        offset = 2
    elif comf_class == 2:
        offset = 3
    elif comf_class == 3:
        offset = 4
    else:
        raise ValueError('Comfort class {} is not an acceptable value. '
                         'Choose from: 1, 2, 3'.format(comf_class))
    return offset


def weighted_running_mean_hourly(outdoor_temperatures, alpha=0.8):
    """Get weighted running mean temperatures given hourly outdoor temperatures.

    Note:
        [1] Nicol, F. and McCartney, K. (2001) Final Report (Public) Smart Controls
        and Thermal Comfort (SCATs). Report to the European Comission of the Smart
        Controls and Thermal Comfort project. Oxford: Oxford Brokes University.

    Args:
        outdoor_temperatures: A list of hourly outdoor temperatures in Celsius for which
            running mean values will be computed.  The list should contain at least
            168 values (1 week of data) in order to be meaningful.  Lists shorter
            than 24 are not acceptable.
        alpha: A constant between 0 and 1 that governs how quickly the running mean
            responds to the outdoor temperature. Default is 0.8, which was found to
            be most suitable by Nicol and McCartney[1].

    Returns:
        prevailing_temp -- A list of prevailing outdoor temperatures with a length that
        matches the input outdoor_temperatures.
    """
    # ensure that there are at least a day's worth of values
    assert len(outdoor_temperatures) >= 168, 'outdoor_temperatures must be for '\
        'at least a week (168 values). Got {} values.'.format(len(outdoor_temperatures))

    # compute the initial prevailing outdoor temperature by looking over the past week
    divisor = 1 + alpha + alpha ** 2 + alpha ** 3 + alpha ** 4 + alpha ** 5
    dividend = (sum(outdoor_temperatures[-24:]) / 24) + \
        (alpha * (sum(outdoor_temperatures[-48:-24]) / 24)) + \
        (alpha ** 2 * (sum(outdoor_temperatures[-72:-48]) / 24)) + \
        (alpha ** 3 * (sum(outdoor_temperatures[-96:-72]) / 24)) + \
        (alpha ** 4 * (sum(outdoor_temperatures[-120:-96]) / 24)) + \
        (alpha ** 5 * (sum(outdoor_temperatures[-144:-120]) / 24))
    starting_temp = dividend / divisor
    daily_run_means = [starting_temp]
    daily_means = [sum(outdoor_temperatures[:24]) / 24]
    prevailing_temp = [starting_temp] * 24

    # run through each day of data and compute the running mean using the previous day's
    start_hour = 24
    for i in xrange(int(math.floor(len(outdoor_temperatures) / 24) - 1)):
        daily_mean = sum(outdoor_temperatures[start_hour: start_hour + 24]) / 24
        daily_run_mean = ((1 - alpha) * daily_means[-1]) + alpha * daily_run_means[-1]
        daily_run_means.append(daily_run_mean)
        daily_means.append(daily_mean)
        prevailing_temp.extend([daily_run_mean] * 24)
        start_hour += 24

    # if there are extra hours, compute the new running mean of the last day.
    if len(prevailing_temp) != len(outdoor_temperatures):
        daily_run_mean = ((1 - alpha) * daily_means[-1]) + alpha * daily_run_means[-1]
        num_extra = len(outdoor_temperatures) - len(prevailing_temp)
        prevailing_temp.extend([daily_run_mean] * num_extra)

    return prevailing_temp


def weighted_running_mean_daily(outdoor_temperatures, alpha=0.8):
    """Get weighted running mean temperatures given average daily outdoor temperatures.

    Note:
        [1] Nicol, F. and McCartney, K. (2001) Final Report (Public) Smart Controls
        and Thermal Comfort (SCATs). Report to the European Comission of the Smart
        Controls and Thermal Comfort project. Oxford: Oxford Brokes University.

    Args:
        outdoor_temperatures: A list of daily outdoor temperatures in Celsius for which
            running mean values will be computed.  The list should contain at least
            7 values in order to be meaningful.  Lists shorter than 1 are not acceptable.
        alpha: A constant between 0 and 1 that governs how quickly the running mean
            responds to the outdoor temperature. Default is 0.8, which was found to
            be most suitable by Nicol and McCartney[1].

    Returns:
        daily_run_means -- A list of prevailing outdoor temperatures with a length that
        matches the input outdoor_temperatures.
    """
    # ensure that there are at least one value
    assert len(outdoor_temperatures) >= 7, 'outdoor_temperatures must have '\
        'at least 7 values to be meaningful.'

    # compute the initial prevailing outdoor temperature by looking over the past week
    divisor = 1 + alpha + alpha ** 2 + alpha ** 3 + alpha ** 4 + alpha ** 5
    dividend = outdoor_temperatures[-1] + alpha * outdoor_temperatures[-2] + \
        alpha ** 2 * outdoor_temperatures[-3] + alpha ** 3 * outdoor_temperatures[-4] + \
        alpha ** 4 * outdoor_temperatures[-5] + alpha ** 5 * outdoor_temperatures[-6]
    starting_temp = dividend / divisor
    daily_run_means = [starting_temp]
    daily_means = [outdoor_temperatures[0]]

    # run through each day of data and compute the running mean using the previous day's
    for i in xrange(len(outdoor_temperatures) - 1):
        daily_run_mean = ((1 - alpha) * daily_means[-1]) + alpha * daily_run_means[-1]
        daily_run_means.append(daily_run_mean)
        daily_means.append(outdoor_temperatures[i])

    return daily_run_means


def check_prevailing_temperatures_ashrae55(t_prevail):
    """Check whether prevailing temperatures are outside permissable ranges for ASHRAE-55.

    Args:
        t_prevail: A list of prevailing outdoor temperature [C].

    Returns:
        A tuple with two elements

        -   all_in_range: A boolean to note whether all of the input t_prevail values
            are in acceptable ranges.
        -   message: Text indicating the number of values that are above and below
            acceptable ranges.
    """
    all_in_range, message = check_prevailing_temperatures_range(
        t_prevail, 10., 33.5, 'ASHRAE-55')
    return all_in_range, message


def check_prevailing_temperatures_en15251(t_prevail):
    """Check whether prevailing temperatures are outside permissable ranges for EN-15251.

    Args:
        t_prevail: A list of prevailing outdoor temperature [C].

    Returns:
        A tuple with two elements

        -   all_in_range: A boolean to note whether all of the input t_prevail values
            are in acceptable ranges.
        -   message: Text indicating the number of values that are above and below
            acceptable ranges.
    """
    all_in_range, message = check_prevailing_temperatures_range(
        t_prevail, 10., 30., 'EN-15251')
    return all_in_range, message


def check_prevailing_temperatures_range(t_prevail, lower, upper, standard=''):
    """Check prevailing temperatures to see how many are outside a certain range.

    Args:
        t_prevail: A list of prevailing outdoor temperature [C].
        lower: A lower limit for the temperatures [C].
        upper: An upper limit for the temperatures [C].
        standard: Optional text for a standard name that will be used in the
            output message (eg. ASHRAE-55).

    Returns:
        A tuple with two elements

        -   all_in_range: A boolean to note whether all of the input t_prevail values
            are in acceptable ranges.
        -   message: Text indicating the number of values that are above and below
            acceptable ranges.
    """
    # defaults to be changed
    all_in_range = True
    num_cold = 0
    num_hot = 0

    # check all values
    for t in t_prevail:
        if t < lower:
            all_in_range = False
            num_cold += 1
        elif t > upper:
            all_in_range = False
            num_hot += 1

    # build the message
    if all_in_range is True:
        message = 'All prevailing temperatures are within acceptable ASHRAE-55 ranges.'
    else:
        cold_msg = '{} prevailing temperatures are colder than the lower limit' \
            ' permitted by {} ({} C).'.format(num_cold, standard, lower)
        hot_msg = '{} prevailing temperatures are hotter than the upper limit' \
            ' permitted by {} ({} C).'.format(num_hot, standard, upper)
        if num_hot == 0:
            message = cold_msg
        elif num_cold == 0:
            message = hot_msg
        else:
            message = '{}\n{}'.format(cold_msg, hot_msg)

    return all_in_range, message



# coding=utf-8
"""Object for calculating UTCI comfort from DataCollections."""
from __future__ import division

from ..utci import universal_thermal_climate_index
from ..parameter.utci import UTCIParameter
from .base import ComfortCollection
from .solarcal import OutdoorSolarCal

from ladybug._datacollectionbase import BaseCollection

from ladybug.datatype.temperature import Temperature, MeanRadiantTemperature, \
    AirTemperature, UniversalThermalClimateIndex
from ladybug.datatype.fraction import RelativeHumidity
from ladybug.datatype.speed import Speed, WindSpeed
from ladybug.datatype.thermalcondition import ThermalComfort, ThermalCondition, \
    ThermalConditionFivePoint, ThermalConditionSevenPoint, \
    ThermalConditionNinePoint, ThermalConditionElevenPoint, UTCICategory


class UTCI(ComfortCollection):
    """UTCI comfort DataCollection object.

    Args:
        air_temperature: Data Collection of air temperature values in Celsius.
        rel_humidity: Data Collection of relative humidity values in % or a
            single relative humidity value to be used for the whole analysis.
        rad_temperature: Data Collection of mean radiant temperature (MRT)
            values in degrees Celsius or a single MRT value to be used for the whole
            analysis. If None, this will be the same as the air_temperature.
        wind_speed: Data Collection of meteorological wind speed values in m/s
            (measured 10 m above the ground) or a single wind speed value to be
            used for the whole analysis. If None, this will default to a very
            low wind speed of 0.1 m/s.
        comfort_parameter: Optional UTCIParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume comfort thresholds consistent with those used by meterologists
            to categorize outdoor conditions.

    Properties:
        * air_temperature
        * rad_temperature
        * air_speed
        * rel_humidity
        * comfort_parameter
        * universal_thermal_climate_index
        * is_comfortable
        * thermal_condition
        * thermal_condition_five_point
        * thermal_condition_seven_point
        * thermal_condition_nine_point
        * thermal_condition_eleven_point
        * original_utci_category
        * percent_comfortable
        * percent_uncomfortable
        * percent_neutral
        * percent_hot
        * percent_cold
        * percent_extreme_cold_stress
        * percent_very_strong_cold_stress
        * percent_strong_cold_stress
        * percent_moderate_cold_stress
        * percent_slight_cold_stress
        * percent_slight_heat_stress
        * percent_moderate_heat_stress
        * percent_strong_heat_stress
        * percent_very_strong_heat_stress
        * percent_extreme_heat_stress
    """
    _model = 'Universal Thermal Climate Index'
    __slots__ = ('_air_temperature', '_rel_humidity', '_rad_temperature', '_wind_speed',
                 '_comfort_par', '_utci', '_thermal_category', '_air_temperature_coll',
                 '_rel_humidity_coll', '_rad_temperature_coll', '_wind_speed_coll',
                 '_utci_coll', '_is_comfortable_coll', '_thermal_condition_coll',
                 '_five_point_coll', '_seven_point_coll', '_nine_point_coll',
                 '_eleven_point_coll', '_original_category_coll')

    def __init__(self, air_temperature, rel_humidity, rad_temperature=None,
                 wind_speed=None, comfort_parameter=None):
        """Initialize a UTCI comfort object from DataCollections of UTCI inputs.
        """
        # set up the object using air temperature as a base
        self._check_datacoll(air_temperature, Temperature, 'C', 'air_temperature')
        self._input_collections = [air_temperature]
        self._calc_length = len(air_temperature.values)
        self._base_collection = air_temperature

        # check required inputs
        self._air_temperature = air_temperature.values
        self._rel_humidity = self._check_input(
            rel_humidity, RelativeHumidity, '%', 'rel_humidity')

        # check inputs with defaults
        if rad_temperature is not None:
            self._rad_temperature = self._check_input(
                rad_temperature, Temperature, 'C', 'rad_temperature')
        else:
            self._rad_temperature = self._air_temperature

        if wind_speed is not None:
            self._wind_speed = self._check_input(
                wind_speed, Speed, 'm/s', 'air_speed')
        else:
            self._wind_speed = [0.1] * self.calc_length

        # check that all input data collections are aligned.
        BaseCollection.are_collections_aligned(self._input_collections)

        # check comfort parameters
        if comfort_parameter is None:
            self._comfort_par = UTCIParameter()
        else:
            assert isinstance(comfort_parameter, UTCIParameter), 'comfort_parameter '\
                'must be a UTCIParameter object. Got {}'.format(type(comfort_parameter))
            self._comfort_par = comfort_parameter

        # compute UTCI
        self._calculate_utci()

    @classmethod
    def from_epw(cls, epw, include_wind=True, include_sun=True,
                 utci_parameter=None):
        """Get a UTCI comfort object from the conditions within an EPW file.

        Args:
            epw: A ladybug EPW object from which the UTCI object will be created.
            include_wind: Set to True to include the EPW wind speed in the calculation.
                Setting to False will assume a condition that is shielded from wind
                where the human experiences a very low wind speed of 0.1 m/s.
                Default is True to include wind.
            include_sun: Set to True to include the mean radiant temperature (MRT) delta
                from both shortwave solar falling directly on people and long wave
                radiant exchange with the sky. Setting to False will assume a shaded
                condition with MRT being equal to the EPW dry bulb temperature. When
                set to True, this calculation will assume no surrounding shade context,
                standing human geometry, and a solar horizontal angle relative to
                front of person (SHARP) of 135 degrees. A SHARP of 135 essentially
                assumes that a person typically faces their side or back to the sun
                to avoid glare. Default: True to include sun.
            utci_parameter: Optional UTCIParameter object to specify parameters under
                which conditions are considered acceptable. If None, default will
                assume comfort thresholds consistent with those used by meterologists
                to categorize outdoor conditions.

        Returns:
            A UTCI object with data collections of the results as properties.

        Usage:

        .. code-block:: python

            from ladybug.epw import EPW
            from ladybug_comfort.collection.utci import UTCI

            epw_file_path = './tests/epw/chicago.epw'
            epw = EPW(epw_file_path)
            utci_exposed = UTCI.from_epw(epw, include_wind=True, include_sun=True)
            utci_protected = UTCI.from_epw(epw, include_wind=False, include_sun=False)

            print(utci_exposed.percent_neutral)  # comfortable % with sun + wind
            print(utci_protected.percent_neutral)  # comfortable % without sun + wind
        """
        # Get wind and mrt inputs
        wind_speed = epw.wind_speed if include_wind is True else 0.1
        if include_sun is True:
            solarcal_obj = OutdoorSolarCal(epw.location, epw.direct_normal_radiation,
                                           epw.diffuse_horizontal_radiation,
                                           epw.horizontal_infrared_radiation_intensity,
                                           epw.dry_bulb_temperature)
            mrt = solarcal_obj.mean_radiant_temperature
        else:
            mrt = epw.dry_bulb_temperature

        # return the comfort object
        return cls(epw.dry_bulb_temperature, epw.relative_humidity, mrt, wind_speed,
                   utci_parameter)

    def _calculate_utci(self):
        """Compute UTCI for each step of the Data Collection."""
        self._utci = []
        self._thermal_category = []
        for ta, tr, vel, rh in \
            zip(self._air_temperature, self._rad_temperature,
                self._wind_speed, self._rel_humidity):
            result = universal_thermal_climate_index(ta, tr, vel, rh)
            self._utci.append(result)
            self._thermal_category.append(
                self._comfort_par.thermal_condition_eleven_point(result))

    @property
    def air_temperature(self):
        """Data Collection of air temperature values in degrees C."""
        return self._get_coll('_air_temperature_coll', self._air_temperature,
                              AirTemperature, 'C')

    @property
    def rad_temperature(self):
        """Data Collection of mean radiant temperature (MRT) values in degrees C."""
        return self._get_coll('_rad_temperature_coll', self._rad_temperature,
                              MeanRadiantTemperature, 'C')

    @property
    def wind_speed(self):
        """Data Collection of air speed values in m/s."""
        return self._get_coll('_wind_speed_coll', self._wind_speed,
                              WindSpeed, 'm/s')

    @property
    def rel_humidity(self):
        """Data Collection of relative humidity values in %."""
        return self._get_coll('_rel_humidity_coll', self._rel_humidity,
                              RelativeHumidity, '%')

    @property
    def comfort_parameter(self):
        """UTCI comfort parameters that are assigned to this object."""
        return self._comfort_par

    @property
    def universal_thermal_climate_index(self):
        """A Data Collection of Universal Thermal Climate Index (UTCI) in C."""
        return self._get_coll('_utci_coll', self._utci,
                              UniversalThermalClimateIndex, 'C')

    @property
    def is_comfortable(self):
        """Data Collection of integers noting whether the input conditions are
        acceptable according to the assigned comfort_parameter.

        Values are one of the following:

        * 0 = uncomfortable
        * 1 = comfortable
        """
        return self._get_coll('_is_comfortable_coll', self._comf_val_funct,
                              ThermalComfort, 'condition')

    @property
    def thermal_condition(self):
        """Data Collection of integers noting the thermal status of a subject
        according to the assigned comfort_parameter.

        Values are one of the following:

        * -1 = cold
        * 0 = netural
        * +1 = hot
        """
        return self._get_coll('_thermal_condition_coll', self._condit_val_funct,
                              ThermalCondition, 'condition')

    @property
    def thermal_condition_five_point(self):
        """Data Collection of integers noting the thermal status on a five-point scale.

        Values are one of the following:

        * -2 = strong/extreme cold stress
        * -1 = moderate cold stress
        * 0 = no thermal stress
        * +1 = moderate heat stress
        * +2 = strong/extreme heat stress
        """
        return self._get_coll('_five_point_coll', self._five_pt_funct,
                              ThermalConditionFivePoint, 'condition')

    @property
    def thermal_condition_seven_point(self):
        """Data Collection of integers noting the thermal status on a seven-point scale.

        Values are one of the following:

        * -3 = very strong/extreme cold stress
        * -2 = strong cold stress
        * -1 = moderate cold stress
        * 0 = no thermal stress
        * +1 = moderate heat stress
        * +2 = strong heat stress
        * +3 = very strong/extreme heat stress
        """
        return self._get_coll('_seven_point_coll', self._seven_pt_funct,
                              ThermalConditionSevenPoint, 'condition')

    @property
    def thermal_condition_nine_point(self):
        """Data Collection of integers noting the thermal status on a nine-point scale.

        Values are one of the following:

        * -4 = very strong/extreme cold stress
        * -3 = strong cold stress
        * -2 = moderate cold stress
        * -1 = slight cold stress
        * 0 = no thermal stress
        * +1 = slight heat stress
        * +2 = moderate heat stress
        * +3 = strong heat stress
        * +4 = very strong/extreme heat stress
        """
        return self._get_coll('_nine_point_coll', self._nine_pt_funct,
                              ThermalConditionNinePoint, 'condition')

    @property
    def thermal_condition_eleven_point(self):
        """Data Collection of integers noting the thermal status on an eleven-point scale.

        Values are one of the following:

        * -5 = extreme cold stress
        * -4 = very strong cold stress
        * -3 = strong cold stress
        * -2 = moderate cold stress
        * -1 = slight cold stress
        * 0 = no thermal stress
        * +1 = slight heat stress
        * +2 = moderate heat stress
        * +3 = strong heat stress
        * +4 = very strong heat stress
        * +5 = extreme heat stress
        """
        return self._get_coll('_eleven_point_coll', self._thermal_category,
                              ThermalConditionElevenPoint, 'condition')

    @property
    def original_utci_category(self):
        """Data Collection of integers noting the original UTCI assessment scale.

        Glossary of Terms for Thermal Physiology (2003).
        Journal of Thermal Biology 28, 75-106

        Values are one of the following:

        * 0 = extreme cold stress
        * 1 = very strong cold stress
        * 2 = strong cold stress
        * 3 = moderate cold stress
        * 4 = slight cold stress
        * 5 = no thermal stress
        * 6 = moderate heat stress
        * 7 = strong heat stress
        * 8 = strong heat stress
        * 9 = extreme heat stress
        """
        return self._get_coll('_original_category_coll', self._original_category_funct,
                              UTCICategory, 'condition')

    @property
    def percent_comfortable(self):
        """The percent of time comfortabe given by the assigned comfort_parameter."""
        _vals = [1 for t in self._thermal_category if t == 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_uncomfortable(self):
        """The percent of time uncomfortable given by the assigned comfort_parameter."""
        return 100 - self.percent_comfortable

    @property
    def percent_neutral(self):
        """The percent of time that the thermal_condition is neutral."""
        return self.percent_comfortable

    @property
    def percent_cold(self):
        """The percent of time that the thermal_condition is cold."""
        _vals = [1 for x in self._thermal_category if x < 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_hot(self):
        """The percent of time that the thermal_condition is hot."""
        _vals = [1 for x in self._thermal_category if x > 0]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_slight_cold_stress(self):
        """The percent of time that conditions have slight cold stress."""
        _vals = [1 for x in self._thermal_category if x == -1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_moderate_cold_stress(self):
        """The percent of time that conditions have moderate cold stress."""
        _vals = [1 for x in self._thermal_category if x == -2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_strong_cold_stress(self):
        """The percent of time that conditions have strong cold stress."""
        _vals = [1 for x in self._thermal_category if x == -3]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_very_strong_cold_stress(self):
        """The percent of time that conditions have very strong cold stress."""
        _vals = [1 for x in self._thermal_category if x == -4]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_extreme_cold_stress(self):
        """The percent of time that conditions have very strong cold stress."""
        _vals = [1 for x in self._thermal_category if x == -5]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_slight_heat_stress(self):
        """The percent of time that conditions have slight heat stress."""
        _vals = [1 for x in self._thermal_category if x == 1]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_moderate_heat_stress(self):
        """The percent of time that conditions have moderate heat stress."""
        _vals = [1 for x in self._thermal_category if x == 2]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_strong_heat_stress(self):
        """The percent of time that conditions have strong heat stress."""
        _vals = [1 for x in self._thermal_category if x == 3]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_very_strong_heat_stress(self):
        """The percent of time that conditions have very strong heat stress."""
        _vals = [1 for x in self._thermal_category if x == 4]
        return (sum(_vals) / self._calc_length) * 100

    @property
    def percent_extreme_heat_stress(self):
        """The percent of time that conditions have very strong heat stress."""
        _vals = [1 for x in self._thermal_category if x == 5]
        return (sum(_vals) / self._calc_length) * 100

    def _comf_val_funct(self):
        return [self._comfort_par.is_comfortable(t) for t in self._utci]

    def _condit_val_funct(self):
        return [self._comfort_par.thermal_condition(t) for t in self._utci]

    def _five_pt_funct(self):
        return [self._comfort_par.thermal_condition_five_point(t) for t in self._utci]

    def _seven_pt_funct(self):
        return [self._comfort_par.thermal_condition_seven_point(t) for t in self._utci]

    def _nine_pt_funct(self):
        return [self._comfort_par.thermal_condition_nine_point(t) for t in self._utci]

    def _original_category_funct(self):
        return [self._comfort_par.original_utci_category(t) for t in self._utci]



# coding utf-8
import pytest

from ladybug_comfort.wc import windchill_index
from ladybug_comfort.wc import windchill_index_effect_category
from ladybug_comfort.wc import windchill_temp
from ladybug_comfort.wc import windchill_temp_effect_category


def test_windchill_index():
    """Test the windchill_index function"""
    assert windchill_index(32, 85) == pytest.approx(20.5216, rel=1e-3)
    assert windchill_index(20, 50) == pytest.approx(471.1182, rel=1e-3)


def test_windchill_index_effect_category():
    """Test the windchill_index_effect_category function"""
    assert windchill_index_effect_category(2350) == -4
    assert windchill_index_effect_category(1700) == -3
    assert windchill_index_effect_category(1000) == -2
    assert windchill_index_effect_category(600) == -1
    assert windchill_index_effect_category(400) == 0
    assert windchill_index_effect_category(150) == 1
    assert windchill_index_effect_category(90) == 2
    assert windchill_index_effect_category(30) == 3


def test_windchill_temp():
    """Test the windchill_temp function"""
    assert windchill_temp(32, 85) == pytest.approx(36.3012, rel=1e-3)
    assert windchill_temp(20, 50) == pytest.approx(17.6540, rel=1e-3)


def test_windchill_temp_effect_category():
    """Test the windchill_index_effect_category function"""
    assert windchill_temp_effect_category(10) == 0
    assert windchill_temp_effect_category(-5) == -1
    assert windchill_temp_effect_category(-15) == -2
    assert windchill_temp_effect_category(-31) == -3
    assert windchill_temp_effect_category(-45) == -4
    assert windchill_temp_effect_category(-50) == -5
    assert windchill_temp_effect_category(-60) == -6



# coding=utf-8
"""Comfort parameter base object."""


class ComfortParameter(object):
    """Thermal comfort parameter base class."""
    _model = None
    __slots__ = ()

    @property
    def comfort_model(self):
        """Return the name of the comfort model to which the parameters belong."""
        return self._model

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()



# coding utf-8
import pytest

from ladybug_comfort.ts import thermal_sensation
from ladybug_comfort.ts import thermal_sensation_effect_category


def test_thermal_sensation():
    """Test the thermal_sensation function."""
    assert thermal_sensation(40, 10, 15, 10, 11) == pytest.approx(2.9208,
                                                                  rel=1e-3)
    assert thermal_sensation(20, 50, 34, 48, 32) == pytest.approx(-12.1482,
                                                                  rel=1e-3)


def test_thermal_sensation_effect_category():
    """Test the thermal_sensation_effect_category function"""
    assert thermal_sensation_effect_category(7.5) == 3
    assert thermal_sensation_effect_category(6.4) == 2
    assert thermal_sensation_effect_category(5.2) == 1
    assert thermal_sensation_effect_category(4.8) == 0
    assert thermal_sensation_effect_category(3.7) == -1
    assert thermal_sensation_effect_category(2.5) == -2
    assert thermal_sensation_effect_category(-4) == -3



# coding=utf-8
"""Objects for calculating solar-adjusted MRT from DataCollections."""
from __future__ import division

from ..solarcal import outdoor_sky_heat_exch, indoor_sky_heat_exch, \
    shortwave_from_horiz_solar, sharp_from_solar_and_body_azimuth
from ..parameter.solarcal import SolarCalParameter
from .base import ComfortCollection

from ladybug.location import Location
from ladybug.sunpath import Sunpath
from ladybug.datacollection import HourlyDiscontinuousCollection

from ladybug.datatype.temperature import Temperature, MeanRadiantTemperature
from ladybug.datatype.temperaturedelta import RadiantTemperatureDelta
from ladybug.datatype.energyflux import Irradiance, EffectiveRadiantField, \
    HorizontalInfraredRadiationIntensity, DiffuseHorizontalIrradiance, \
    DirectNormalIrradiance, DirectHorizontalIrradiance
from ladybug.datatype.energyintensity import Radiation
from ladybug.datatype.fraction import Fraction


class _SolarCalBase(ComfortCollection):
    """Base class used by all objects that use SolarCal with Data Collections."""
    __slots__ = ('_location', '_fract_exp', '_flr_ref', '_body_par', '_dmrt', '_mrt',
                 '_fract_exp_coll', '_flr_ref_coll', '_dmrt_coll', '_mrt_coll')

    def __init__(self, location, fraction_body_exposed=None, floor_reflectance=None,
                 solarcal_body_parameter=None):
        # set required inputs
        assert isinstance(location, Location), 'location must be a Ladybug Location' \
            ' object. Got {}.'.format(type(location))
        self._location = location.duplicate()

        # check optional inputs
        self._fract_exp = self._fraction_input_check(
            fraction_body_exposed, 'fraction_body_exposed', 1)
        self._flr_ref = self._fraction_input_check(
            floor_reflectance, 'floor_reflectance', 0.25)

        # check comfort parameters
        self._body_par_check(solarcal_body_parameter)

        # setup lists to be filled
        self._dmrt = []
        self._mrt = []

    @property
    def location(self):
        """Ladybug Location object."""
        return self._location.duplicate()

    @property
    def fraction_body_exposed(self):
        """Data Collection of body fraction exposed to direct sun."""
        return self._get_coll('_fract_exp_coll', self._fract_exp,
                              Fraction('Fraction Body Exposed'), 'fraction')

    @property
    def floor_reflectance(self):
        """Data Collection of floor reflectance."""
        return self._get_coll('_flr_ref_coll', self._flr_ref,
                              Fraction('Floor Reflectance'), 'fraction')

    @property
    def solarcal_body_parameter(self):
        """SolarCal body parameters that are assigned to this object."""
        return self._body_par

    @property
    def mrt_delta(self):
        """Data Collection of total MRT delta in C."""
        return self._get_coll('_dmrt_coll', self._dmrt, RadiantTemperatureDelta, 'dC')

    @property
    def mean_radiant_temperature(self):
        """Data Collection of total mean radiant temperature in C."""
        return self._get_coll('_mrt_coll', self._mrt, MeanRadiantTemperature, 'C')

    def _radiation_check(self, data_coll, name):
        assert isinstance(data_coll, HourlyDiscontinuousCollection), \
            '{} must be an hourly collection. Got {}.'.format(name, type(data_coll))
        if isinstance(data_coll.header.data_type, Radiation):
            self._check_datacoll(data_coll, Radiation, 'Wh/m2', name)
            timestep = data_coll.header.analysis_period.timestep
            assert timestep == 1, '{} timestep must be 1 when using Radiation as the ' \
                'data type. Got timestep of {}'.format(name, timestep)
        else:
            self._check_datacoll(data_coll, Irradiance, 'W/m2', name)

    def _fraction_input_check(self, data_coll, name, default):
        if data_coll is not None:
            return self._check_input(data_coll, Fraction, 'fraction', name)
        else:
            return [default] * self._calc_length

    def _body_par_check(self, body_par):
        if body_par is None:
            self._body_par = SolarCalParameter(posture='standing')
        else:
            assert isinstance(body_par, SolarCalParameter), \
                'solarcal_body_parameter must be a SolarCalParameter object. Got {}'\
                .format(type(body_par))
            self._body_par = body_par

    def _get_altitudes_and_sharps(self):
        """Get altitudes and sharps from solar position."""
        sp = Sunpath.from_location(self._location)
        _altitudes = []
        if self._body_par.body_azimuth is None:
            _sharps = [self._body_par.sharp] * self._calc_length
            for t_date in self._base_collection.datetimes:
                sun = sp.calculate_sun_from_date_time(t_date)
                _altitudes.append(sun.altitude)
        else:
            _sharps = []
            for t_date in self._base_collection.datetimes:
                sun = sp.calculate_sun_from_date_time(t_date)
                sharp = sharp_from_solar_and_body_azimuth(sun.azimuth,
                                                          self._body_par.body_azimuth)
                _sharps.append(sharp)
                _altitudes.append(sun.altitude)
        return _altitudes, _sharps


class OutdoorSolarCal(_SolarCalBase):
    """Outdoor SolarCal Collection object.

    Args:
        location: A Ladybug Location object.
        direct_normal_solar: Hourly Data Collection with the direct normal solar
            irradiance in W/m2.
        diffuse_horizontal_solar: Hourly Data Collection with the diffuse
            horizontal solar irradiance in W/m2.
        surface_temperatures: Hourly Data Collection with the temperature of surfaces
            around the person in degrees Celsius. This includes the ground and
            any other surfaces blocking the view to the sky. When the temperature
            of these individual surfaces are known, the input here should be the
            average temperature of the surfaces weighted by view-factor to the human.
            When such individual surface temperatures are unknown, the outdoor
            dry bulb temperature is typically used as a proxy.
        horizontal_infrared: Hourly Data Collection with the horizontal infrared
            radiation intensity from the sky in W/m2.
        fraction_body_exposed: A Data Collection or number between 0 and 1
            representing the fraction of the body exposed to direct sunlight.
            Note that this does not include the body’s self-shading; only the
            shading from surroundings.
            Default is 1 for a person standing in an open area.
        sky_exposure: A Data Collection or number between 0 and 1 representing the
            fraction of the sky vault in occupant’s view. Default is 1 for a person
            standing in an open area.
        floor_reflectance: A Data Collection or number between 0 and 1 that
            represents the reflectance of the floor. Default is for 0.25 which
            is characteristic of outdoor grass or dry bare soil.
        solarcal_body_parameter: Optional SolarCalParameter object to account for
            properties of the human geometry.

    Properties:
        * location
        * direct_normal_solar
        * diffuse_horizontal_solar
        * horizontal_infrared
        * surface_temperatures
        * fraction_body_exposed
        * sky_exposure
        * floor_reflectance
        * solarcal_body_parameter
        * shortwave_effective_radiant_field
        * longwave_effective_radiant_field
        * shortwave_mrt_delta
        * longwave_mrt_delta
        * mrt_delta
        * mean_radiant_temperature
    """
    _model = 'Outdoor SolarCal'
    __slots__ = ('_dir_norm', '_diff_horiz', '_horiz_ir', '_srf_temp', '_sky_exp',
                 '_s_erf', '_s_dmrt', '_l_erf', '_l_dmrt', '_dir_norm_coll',
                 '_diff_horiz_coll', '_horiz_ir_coll', '_srf_temp_coll', '_sky_exp_coll',
                 '_s_erf_coll', '_s_dmrt_coll', '_l_erf_coll', '_l_dmrt_coll')

    def __init__(self, location, direct_normal_solar, diffuse_horizontal_solar,
                 horizontal_infrared, surface_temperatures,
                 fraction_body_exposed=None, sky_exposure=None,
                 floor_reflectance=None, solarcal_body_parameter=None):
        """Initialize Outdoor SolarCal object.
        """
        # set up the object using radiation as a base
        self._radiation_check(direct_normal_solar, 'direct_normal_solar')
        self._radiation_check(diffuse_horizontal_solar, 'diffuse_horizontal_solar')
        self._input_collections = [direct_normal_solar, diffuse_horizontal_solar]
        self._calc_length = len(direct_normal_solar.values)
        self._base_collection = direct_normal_solar

        # check required inputs
        _SolarCalBase.__init__(self, location, fraction_body_exposed, floor_reflectance,
                               solarcal_body_parameter)
        self._dir_norm = direct_normal_solar.values
        self._diff_horiz = diffuse_horizontal_solar.values
        self._horiz_ir = self._check_input(
            horizontal_infrared, HorizontalInfraredRadiationIntensity, 'W/m2',
            'horizontal_infrared')
        self._srf_temp = self._check_input(
            surface_temperatures, Temperature, 'C', 'surface_temperatures')

        # check optional inputs
        self._sky_exp = self._fraction_input_check(sky_exposure, 'sky_exposure', 1)

        # check that all input data collections are aligned.
        HourlyDiscontinuousCollection.are_collections_aligned(self._input_collections)

        # compute SolarCal
        self._calculate_solarcal()

    def _calculate_solarcal(self):
        """Compute SolarCal for each step of the Data Collection."""
        # empty lists to be filled
        self._s_erf = []
        self._s_dmrt = []
        self._l_erf = []
        self._l_dmrt = []

        # get altitudes and sharps from solar position
        _altitudes, _sharps = self._get_altitudes_and_sharps()

        # calculate final erfs and mrt deltas
        for t_srfs, horiz_ir, diff, dir, alt, sharp, sky_e, fract_e, flr_ref in \
                zip(self._srf_temp, self._horiz_ir, self._diff_horiz, self._dir_norm,
                    _altitudes, _sharps, self._sky_exp, self._fract_exp, self._flr_ref):

            result = outdoor_sky_heat_exch(t_srfs, horiz_ir, diff, dir, alt, sky_e,
                                           fract_e, flr_ref, self._body_par.posture,
                                           sharp, self._body_par.body_absorptivity,
                                           self._body_par.body_emissivity)
            self._s_erf.append(result['s_erf'])
            self._s_dmrt.append(result['s_dmrt'])
            self._l_erf.append(result['l_erf'])
            self._l_dmrt.append(result['l_dmrt'])
            self._dmrt.append(result['s_dmrt'] + result['l_dmrt'])
            self._mrt.append(result['mrt'])

    @property
    def diffuse_horizontal_solar(self):
        """Data Collection of diffuse horizontal irradiance in W/m2."""
        return self._get_coll('_diff_horiz_coll', self._diff_horiz,
                              DiffuseHorizontalIrradiance, 'W/m2')

    @property
    def direct_normal_solar(self):
        """Data Collection of direct normal irradiance in W/m2."""
        return self._get_coll('_dir_norm_coll', self._dir_norm,
                              DirectNormalIrradiance, 'W/m2')

    @property
    def surface_temperatures(self):
        """Data Collection of surface temperature values in degrees C."""
        return self._get_coll('_srf_temp_coll', self._srf_temp,
                              Temperature('Surface Temperature'), 'C')

    @property
    def horizontal_infrared(self):
        """Data Collection of horizontal infrared radiation intensity in W/m2."""
        return self._get_coll('_horiz_ir_coll', self._horiz_ir,
                              HorizontalInfraredRadiationIntensity, 'W/m2')

    @property
    def sky_exposure(self):
        """Data Collection of sky view."""
        return self._get_coll('_sky_exp_coll', self._sky_exp,
                              Fraction('Sky Exposure'), 'fraction')

    @property
    def shortwave_effective_radiant_field(self):
        """Data Collection of shortwave effective radiant field in W/m2."""
        return self._get_coll('_s_erf_coll', self._s_erf,
                              EffectiveRadiantField, 'W/m2')

    @property
    def longwave_effective_radiant_field(self):
        """Data Collection of longwave effective radiant field in W/m2."""
        return self._get_coll('_l_erf_coll', self._l_erf,
                              EffectiveRadiantField, 'W/m2')

    @property
    def shortwave_mrt_delta(self):
        """Data Collection of shortwave MRT delta in C."""
        return self._get_coll('_s_dmrt_coll', self._s_dmrt,
                              RadiantTemperatureDelta, 'dC')

    @property
    def longwave_mrt_delta(self):
        """Data Collection of longwave MRT delta in C."""
        return self._get_coll('_l_dmrt_coll', self._l_dmrt,
                              RadiantTemperatureDelta, 'dC')


class IndoorSolarCal(_SolarCalBase):
    """Indoor SolarCal Collection object.

    Args:
        location: A Ladybug Location object.
        direct_normal_solar: Hourly Data Collection with the direct normal solar
            irradiance in W/m2.
        diffuse_horizontal_solar: Hourly Data Collection with the diffuse
            horizontal solar irradiance in W/m2.
        longwave_mrt: Hourly Data Collection or individual value with the longwave
            mean radiant temperature (MRT) expereinced as a result of indoor
            surface temperatures in C.
        fraction_body_exposed: A Data Collection or number between 0 and 1
            representing the fraction of the body exposed to direct sunlight.
            Note that this does not include the body’s self-shading; only the
            shading from surroundings.
            Default is 1 for a person standing in an open area.
        sky_exposure: A Data Collection or number between 0 and 1 representing the
            fraction of the sky vault in occupant’s view. Default is 1 for a person
            standing in an open area.
        floor_reflectance: A Data Collection or number between 0 and 1 that
            represents the reflectance of the floor. Default is for 0.25 which
            is characteristic of outdoor grass or dry bare soil.
        window_transmittance: A Data Collection or number between 0 and 1 that
            represents the broadband solar transmittance of the window through which
            the sun is coming. Such values tend to be slightly less than the
            SHGC. Values might be as low as 0.2 and could be as high as 0.85
            for a single pane of glass. Default is 0.4 assuming a double pane
            window with a relatively mild low-e coating.
        solarcal_body_parameter: Optional SolarCalParameter object to account for
            properties of the human geometry.

    Properties:
        * location
        * direct_normal_solar
        * diffuse_horizontal_solar
        * longwave_mrt
        * fraction_body_exposed
        * sky_exposure
        * floor_reflectance
        * window_transmittance
        * solarcal_body_parameter
        * effective_radiant_field
        * mrt_delta
        * mean_radiant_temperature
    """
    _model = 'Indoor SolarCal'
    __slots__ = ('_dir_norm', '_diff_horiz', '_l_mrt', '_sky_exp', '_win_trans',
                 '_erf', '_dmrt', '_dir_norm_coll', '_diff_horiz_coll', '_l_mrt_coll',
                 '_sky_exp_coll', '_win_trans_coll', '_erf_coll', '_dmrt_coll')

    def __init__(self, location, direct_normal_solar, diffuse_horizontal_solar,
                 longwave_mrt, fraction_body_exposed=None, sky_exposure=None,
                 floor_reflectance=None, window_transmittance=None,
                 solarcal_body_parameter=None):
        """Initialize Indoor SolarCal object.
        """
        # set up the object using radiation as a base
        self._radiation_check(direct_normal_solar, 'direct_normal_solar')
        self._radiation_check(diffuse_horizontal_solar, 'diffuse_horizontal_solar')
        self._input_collections = [direct_normal_solar, diffuse_horizontal_solar]
        self._calc_length = len(direct_normal_solar.values)
        self._base_collection = direct_normal_solar

        # check required inputs
        _SolarCalBase.__init__(self, location, fraction_body_exposed, floor_reflectance,
                               solarcal_body_parameter)
        self._dir_norm = direct_normal_solar.values
        self._diff_horiz = diffuse_horizontal_solar.values
        self._l_mrt = self._check_input(longwave_mrt, Temperature, 'C', 'longwave_mrt')

        # check optional inputs
        self._sky_exp = self._fraction_input_check(
            sky_exposure, 'sky_exposure', 1)
        self._win_trans = self._fraction_input_check(
            window_transmittance, 'window_transmittance', 0.4)

        # check that all input data collections are aligned.
        HourlyDiscontinuousCollection.are_collections_aligned(self._input_collections)

        # compute SolarCal
        self._calculate_solarcal()

    def _calculate_solarcal(self):
        """Compute SolarCal for each step of the Data Collection."""
        # empty lists to be filled
        self._erf = []

        # get altitudes and sharps from solar position
        _altitudes, _sharps = self._get_altitudes_and_sharps()

        # calculate final erfs and mrt deltas
        for l_mrt, diff, dir, alt, sharp, sky_e, fract_e, flr_ref, w_trans in \
                zip(self._l_mrt, self._diff_horiz, self._dir_norm, _altitudes, _sharps,
                    self._sky_exp, self._fract_exp, self._flr_ref, self._win_trans):

            result = indoor_sky_heat_exch(l_mrt, diff, dir, alt, sky_e, fract_e,
                                          flr_ref, w_trans, self._body_par.posture,
                                          sharp, self._body_par.body_absorptivity,
                                          self._body_par.body_emissivity)
            self._erf.append(result['erf'])
            self._dmrt.append(result['dmrt'])
            self._mrt.append(result['mrt'])

    @property
    def diffuse_horizontal_solar(self):
        """Data Collection of diffuse horizontal irradiance in Wh/m2 or W/m2."""
        return self._get_coll('_diff_horiz_coll', self._diff_horiz,
                              DiffuseHorizontalIrradiance, 'W/m2')

    @property
    def direct_normal_solar(self):
        """Data Collection of direct normal irradiance in Wh/m2 or W/m2."""
        return self._get_coll('_dir_norm_coll', self._dir_norm,
                              DirectNormalIrradiance, 'W/m2')

    @property
    def longwave_mrt(self):
        """Data Collection of surface temperature values in degrees C."""
        return self._get_coll('_l_mrt_coll', self._l_mrt,
                              MeanRadiantTemperature, 'C')

    @property
    def sky_exposure(self):
        """Data Collection of sky view."""
        return self._get_coll('_sky_exp_coll', self._sky_exp,
                              Fraction('Sky Exposure'), 'fraction')

    @property
    def window_transmittance(self):
        """Data Collection of window transmittance."""
        return self._get_coll('_win_trans_coll', self._win_trans,
                              Fraction('Window Transmittance'), 'fraction')

    @property
    def effective_radiant_field(self):
        """Data Collection of shortwave effective radiant field in W/m2."""
        return self._get_coll('_erf_coll', self._erf, EffectiveRadiantField, 'W/m2')

    @property
    def mrt_delta(self):
        """Data Collection of shortwave MRT delta in C."""
        return self._get_coll('_dmrt_coll', self._dmrt, RadiantTemperatureDelta, 'dC')


class HorizontalSolarCal(_SolarCalBase):
    """SolarCal Collection object from horizontal solar components.

    This is particularly useful when trying to estimate solar MRT deltas from
    Radiance radiation simulation result.

    Args:
        location: A Ladybug Location object.
        direct_horizontal_solar: Hourly Data Collection with the direct horizontal
            solar irradiance in W/m2.
        diffuse_horizontal_solar: Hourly Data Collection with the diffuse
            horizontal solar irradiance in W/m2.
        longwave_mrt: Hourly Data Collection or individual value with the longwave
            mean radiant temperature (MRT) expereinced as a result of indoor
            surface temperatures in C.
        fraction_body_exposed: A Data Collection or number between 0 and 1
            representing the fraction of the body exposed to direct sunlight.
            Note that this does not include the body’s self-shading; only the
            shading from surroundings.
            Default is 1 for a person standing in an open area.
        floor_reflectance: A Data Collection or number between 0 and 1 that
            represents the reflectance of the floor. Default is for 0.25 which
            is characteristic of outdoor grass or dry bare soil.
        solarcal_body_parameter: Optional SolarCalParameter object to account for
            properties of the human geometry.

    Properties:
        * location
        * direct_horizontal_solar
        * diffuse_horizontal_solar
        * longwave_mrt
        * fraction_body_exposed
        * floor_reflectance
        * solarcal_body_parameter
        * effective_radiant_field
        * mrt_delta
        * mean_radiant_temperature
    """
    _model = 'Horizontal SolarCal'
    __slots__ = ('_dir_horiz', '_diff_horiz', '_l_mrt', '_erf', '_dmrt',
                 '_dir_horiz_coll', '_diff_horiz_coll', '_l_mrt_coll',
                 '_erf_coll', '_dmrt_coll')

    def __init__(self, location, direct_horizontal_solar, diffuse_horizontal_solar,
                 longwave_mrt, fraction_body_exposed=None,
                 floor_reflectance=None, solarcal_body_parameter=None):
        """Initialize Horizontal SolarCal object.
        """
        # set up the object using radiation as a base
        self._radiation_check(direct_horizontal_solar, 'direct_horizontal_solar')
        self._radiation_check(diffuse_horizontal_solar, 'diffuse_horizontal_solar')
        self._input_collections = [direct_horizontal_solar, diffuse_horizontal_solar]
        self._calc_length = len(direct_horizontal_solar.values)
        self._base_collection = direct_horizontal_solar

        # check required inputs
        _SolarCalBase.__init__(self, location, fraction_body_exposed, floor_reflectance,
                               solarcal_body_parameter)
        self._dir_horiz = direct_horizontal_solar.values
        self._diff_horiz = diffuse_horizontal_solar.values
        self._l_mrt = self._check_input(longwave_mrt, Temperature, 'C', 'longwave_mrt')

        # check that all input data collections are aligned.
        HourlyDiscontinuousCollection.are_collections_aligned(self._input_collections)

        # compute SolarCal
        self._calculate_solarcal()

    def _calculate_solarcal(self):
        """Compute SolarCal for each step of the Data Collection."""
        # empty lists to be filled
        self._erf = []

        # get altitudes and sharps from solar position
        _altitudes, _sharps = self._get_altitudes_and_sharps()

        # calculate final erfs and mrt deltas
        for l_mrt, diff, dir, alt, sharp, fract_e, flr_ref, in \
                zip(self._l_mrt, self._diff_horiz, self._dir_horiz, _altitudes, _sharps,
                    self._fract_exp, self._flr_ref):

            result = shortwave_from_horiz_solar(l_mrt, diff, dir, alt, fract_e,
                                                flr_ref, self._body_par.posture,
                                                sharp, self._body_par.body_absorptivity,
                                                self._body_par.body_emissivity)
            self._erf.append(result['erf'])
            self._dmrt.append(result['dmrt'])
            self._mrt.append(result['mrt'])

    @property
    def diffuse_horizontal_solar(self):
        """Data Collection of diffuse horizontal irradiance in Wh/m2 or W/m2."""
        return self._get_coll('_diff_horiz_coll', self._diff_horiz,
                              DiffuseHorizontalIrradiance, 'W/m2')

    @property
    def direct_horizontal_solar(self):
        """Data Collection of direct horizontal irradiance in Wh/m2 or W/m2."""
        return self._get_coll('_dir_horiz_coll', self._dir_horiz,
                              DirectHorizontalIrradiance, 'W/m2')

    @property
    def longwave_mrt(self):
        """Data Collection of surface temperature values in degrees C."""
        return self._get_coll('_l_mrt_coll', self._l_mrt, MeanRadiantTemperature, 'C')

    @property
    def effective_radiant_field(self):
        """Data Collection of shortwave effective radiant field in W/m2."""
        return self._get_coll('_erf_coll', self._erf, EffectiveRadiantField, 'W/m2')

    @property
    def mrt_delta(self):
        """Data Collection of shortwave MRT delta in C."""
        return self._get_coll('_dmrt_coll', self._dmrt, RadiantTemperatureDelta, 'dC')



# coding=utf-8
"""Utility functions for calculating Thermal Sensation (TS)."""


def thermal_sensation(ta, ws, rh, sr, tground):
    """Calculate Thermal Sensation (TS) from air temperature, wind speed,
    relative humidity, solar radiation and ground temperature.

    Thermal Sensation is an index which predicts sensation of
    satisfaction/dissatisfaction under the prevailing outdoor climatic
    conditions.
    Thermal Sensation is derived from original work carried out by Givoni and
    Noguchi [1].

    Note:
        [1] Givoni, Baruch & Noguchi, Mikiko & Saaroni, Hadas & Potchter,
        Oded & Yaakov, Yaron & Feller, Noa & Becker, Stefan. (2003). Outdoor
        comfort research issues. Energy and Buildings. 35. 77-86.
        10.1016/S0378-7788(02)00082-8.

    Args:
        ta: Air temperature [C]
        ws: Wind speed [m/s]
        rh: Relative humidity [%]
        sr: Solar radiation [Wh/m2]
        tground: Ground temperature [C]

    Returns:
        ts -- Thermal sensation [unitless]
    """
    ts = 1.7 + 0.1118 * ta + 0.0019 * sr - 0.322 * ws - 0.0073 * rh + 0.0054 \
        * tground

    return ts


def thermal_sensation_effect_category(ts):
    """Get the category of effect associated with a given thermal sensation
    (TS).

    Each number (from -3 to 3) represents a certain TS thermal sensation
    category. With categories being the following:

    * -3 = Very cold
    * -2 = Quite cold
    * -1 = Cold
    * 0 = Comfort
    * 1 = Hot
    * 2 = Quite Hot
    * 3 = Very hot

    Args:
        ts: Thermal Sensation [unitless]

    Returns:
        category -- An integer indicating the level of effect associated with the
        thermal sensation. Values are one of the following:

        -   -3 = Very cold
        -   -2 = Quite cold
        -   -1 = Cold
        -    0 = Comfort
        -    1 = Hot
        -    2 = Quite Hot
        -    3 = Very hot
    """
    if ts >= 7:
        category = 3
    elif ts >= 6:
        category = 2
    elif ts >= 5:
        category = 1
    elif ts >= 4:
        category = 0
    elif ts >= 3:
        category = -1
    elif ts >= 2:
        category = -2
    else:
        category = -3

    return category



# coding=utf-8
"""Utility functions for calculating UTCI."""
from __future__ import division

from ladybug.rootfinding import secant
from ladybug.rootfinding import bisect

import math


def universal_thermal_climate_index(ta, tr, vel, rh):
    """Calculate Universal Thermal Climate Index (UTCI) using a polynomial approximation.

    UTCI is an international standard for outdoor temperature sensation
    (aka. "feels-like" temperature) that attempts to fill the
    following requirements:

    1) Thermo-physiological significance in the whole range of heat
       exchange conditions of existing thermal environments
    2) Valid in all climates, seasons, and scales
    3) Useful for key applications in human biometeorology.

    This function here is a Python version of the original UTCI_approx
    application written in Fortran. Version a 0.002, October 2009
    The original Fortran code can be found at www.utci.org.

    Note:
        [1] Peter Bröde, Dusan Fiala, Krzysztof Blazejczyk, Yoram Epstein,
        Ingvar Holmér, Gerd Jendritzky, Bernhard Kampmann, Mark Richards,
        Hannu Rintamäki, Avraham Shitzer, George Havenith. 2009.
        Calculating UTCI Equivalent Temperature. In: JW Castellani & TL
        Endrusick, eds. Proceedings of the 13th International Conference
        on Environmental Ergonomics, USARIEM, Natick, MA.

    Args:
        ta: Air temperature [C]
        tr: Mean radiant temperature [C]
        vel: Wind speed 10 m above ground level [m/s].
            Note that this meteorological speed at 10 m is simply 1.5 times the
            speed felt at ground in the original Fiala model used to build UTCI.
        rh: Relative humidity [%]

    Returns:
        UTCI_approx -- The Universal Thermal Climate Index (UTCI) for the input
        conditions as approximated by a 4-D polynomial.
    """
    # set upper and lower limits of air velocity according to Fiala model scenarios
    vel = 0.5 if vel < 0.5 else vel
    vel = 17 if vel > 17 else vel

    # metrics derived from the inputs used in the polynomial equation
    eh_pa = saturated_vapor_pressure_hpa(ta) * (rh / 100.0)  # partial vapor pressure
    pa_pr = eh_pa / 10.0  # convert vapour pressure to kPa
    d_tr = tr - ta  # difference between radiant and air temperature

    utci_approx = ta + \
        0.607562052 + \
        -0.0227712343 * ta + \
        8.06470249e-4 * ta * ta + \
        -1.54271372e-4 * ta * ta * ta + \
        -3.24651735e-6 * ta * ta * ta * ta + \
        7.32602852e-8 * ta * ta * ta * ta * ta + \
        1.35959073e-9 * ta * ta * ta * ta * ta * ta + \
        -2.25836520 * vel + \
        0.0880326035 * ta * vel + \
        0.00216844454 * ta * ta * vel + \
        -1.53347087e-5 * ta * ta * ta * vel + \
        -5.72983704e-7 * ta * ta * ta * ta * vel + \
        -2.55090145e-9 * ta * ta * ta * ta * ta * vel + \
        -0.751269505 * vel * vel + \
        -0.00408350271 * ta * vel * vel + \
        -5.21670675e-5 * ta * ta * vel * vel + \
        1.94544667e-6 * ta * ta * ta * vel * vel + \
        1.14099531e-8 * ta * ta * ta * ta * vel * vel + \
        0.158137256 * vel * vel * vel + \
        -6.57263143e-5 * ta * vel * vel * vel + \
        2.22697524e-7 * ta * ta * vel * vel * vel + \
        -4.16117031e-8 * ta * ta * ta * vel * vel * vel + \
        -0.0127762753 * vel * vel * vel * vel + \
        9.66891875e-6 * ta * vel * vel * vel * vel + \
        2.52785852e-9 * ta * ta * vel * vel * vel * vel + \
        4.56306672e-4 * vel * vel * vel * vel * vel + \
        -1.74202546e-7 * ta * vel * vel * vel * vel * vel + \
        -5.91491269e-6 * vel * vel * vel * vel * vel * vel + \
        0.398374029 * d_tr + \
        1.83945314e-4 * ta * d_tr + \
        -1.73754510e-4 * ta * ta * d_tr + \
        -7.60781159e-7 * ta * ta * ta * d_tr + \
        3.77830287e-8 * ta * ta * ta * ta * d_tr + \
        5.43079673e-10 * ta * ta * ta * ta * ta * d_tr + \
        -0.0200518269 * vel * d_tr + \
        8.92859837e-4 * ta * vel * d_tr + \
        3.45433048e-6 * ta * ta * vel * d_tr + \
        -3.77925774e-7 * ta * ta * ta * vel * d_tr + \
        -1.69699377e-9 * ta * ta * ta * ta * vel * d_tr + \
        1.69992415e-4 * vel*vel*d_tr + \
        -4.99204314e-5 * ta * vel * vel * d_tr + \
        2.47417178e-7 * ta * ta * vel * vel * d_tr + \
        1.07596466e-8 * ta * ta * ta * vel * vel * d_tr + \
        8.49242932e-5 * vel * vel * vel * d_tr + \
        1.35191328e-6 * ta * vel * vel * vel * d_tr + \
        -6.21531254e-9 * ta * ta * vel * vel * vel * d_tr + \
        -4.99410301e-6 * vel * vel * vel * vel * d_tr + \
        -1.89489258e-8 * ta * vel * vel * vel * vel * d_tr + \
        8.15300114e-8 * vel * vel * vel * vel * vel * d_tr + \
        7.55043090e-4 * d_tr * d_tr + \
        -5.65095215e-5 * ta * d_tr * d_tr + \
        -4.52166564e-7 * ta * ta * d_tr * d_tr + \
        2.46688878e-8 * ta * ta * ta * d_tr * d_tr + \
        2.42674348e-10 * ta * ta * ta * ta * d_tr * d_tr + \
        1.54547250e-4 * vel * d_tr * d_tr + \
        5.24110970e-6 * ta * vel * d_tr * d_tr + \
        -8.75874982e-8 * ta * ta * vel * d_tr * d_tr + \
        -1.50743064e-9 * ta * ta * ta * vel * d_tr * d_tr + \
        -1.56236307e-5 * vel * vel * d_tr * d_tr + \
        -1.33895614e-7 * ta * vel * vel * d_tr * d_tr + \
        2.49709824e-9 * ta * ta * vel * vel * d_tr * d_tr + \
        6.51711721e-7 * vel * vel * vel * d_tr * d_tr + \
        1.94960053e-9 * ta * vel * vel * vel * d_tr * d_tr + \
        -1.00361113e-8 * vel * vel * vel * vel * d_tr * d_tr + \
        -1.21206673e-5 * d_tr * d_tr * d_tr + \
        -2.18203660e-7 * ta * d_tr * d_tr * d_tr + \
        7.51269482e-9 * ta * ta * d_tr * d_tr * d_tr + \
        9.79063848e-11 * ta * ta * ta * d_tr * d_tr * d_tr + \
        1.25006734e-6 * vel * d_tr * d_tr * d_tr + \
        -1.81584736e-9 * ta * vel * d_tr * d_tr * d_tr + \
        -3.52197671e-10 * ta * ta * vel * d_tr * d_tr * d_tr + \
        -3.36514630e-8 * vel * vel * d_tr * d_tr * d_tr + \
        1.35908359e-10 * ta * vel * vel * d_tr * d_tr * d_tr + \
        4.17032620e-10 * vel * vel * vel * d_tr * d_tr * d_tr + \
        -1.30369025e-9 * d_tr * d_tr * d_tr * d_tr + \
        4.13908461e-10 * ta * d_tr * d_tr * d_tr * d_tr + \
        9.22652254e-12 * ta * ta * d_tr * d_tr * d_tr * d_tr + \
        -5.08220384e-9 * vel * d_tr * d_tr * d_tr * d_tr + \
        -2.24730961e-11 * ta * vel * d_tr * d_tr * d_tr * d_tr + \
        1.17139133e-10 * vel * vel * d_tr * d_tr * d_tr * d_tr + \
        6.62154879e-10 * d_tr * d_tr * d_tr * d_tr * d_tr + \
        4.03863260e-13 * ta * d_tr * d_tr * d_tr * d_tr * d_tr + \
        1.95087203e-12 * vel * d_tr * d_tr * d_tr * d_tr * d_tr + \
        -4.73602469e-12 * d_tr * d_tr * d_tr * d_tr * d_tr * d_tr + \
        5.12733497 * pa_pr + \
        -0.312788561 * ta * pa_pr + \
        -0.0196701861 * ta * ta * pa_pr + \
        9.99690870e-4 * ta * ta * ta * pa_pr + \
        9.51738512e-6 * ta * ta * ta * ta * pa_pr + \
        -4.66426341e-7 * ta * ta * ta * ta * ta * pa_pr + \
        0.548050612 * vel * pa_pr + \
        -0.00330552823 * ta * vel * pa_pr + \
        -0.00164119440 * ta * ta * vel * pa_pr + \
        -5.16670694e-6 * ta * ta * ta * vel * pa_pr + \
        9.52692432e-7 * ta * ta * ta * ta * vel * pa_pr + \
        -0.0429223622 * vel * vel * pa_pr + \
        0.00500845667 * ta * vel * vel * pa_pr + \
        1.00601257e-6 * ta * ta * vel * vel * pa_pr + \
        -1.81748644e-6 * ta * ta * ta * vel * vel * pa_pr + \
        -1.25813502e-3 * vel * vel * vel * pa_pr + \
        -1.79330391e-4 * ta * vel * vel * vel * pa_pr + \
        2.34994441e-6 * ta * ta * vel * vel * vel * pa_pr + \
        1.29735808e-4 * vel * vel * vel * vel * pa_pr + \
        1.29064870e-6 * ta * vel * vel * vel * vel * pa_pr + \
        -2.28558686e-6 * vel * vel * vel * vel * vel * pa_pr + \
        -0.0369476348 * d_tr * pa_pr + \
        0.00162325322 * ta * d_tr * pa_pr + \
        -3.14279680e-5 * ta * ta * d_tr * pa_pr + \
        2.59835559e-6 * ta * ta * ta * d_tr * pa_pr + \
        -4.77136523e-8 * ta * ta * ta * ta * d_tr * pa_pr + \
        8.64203390e-3 * vel * d_tr * pa_pr + \
        -6.87405181e-4 * ta * vel * d_tr * pa_pr + \
        -9.13863872e-6 * ta * ta * vel * d_tr * pa_pr + \
        5.15916806e-7 * ta * ta * ta * vel * d_tr * pa_pr + \
        -3.59217476e-5 * vel * vel * d_tr * pa_pr + \
        3.28696511e-5 * ta * vel * vel * d_tr * pa_pr + \
        -7.10542454e-7 * ta * ta * vel * vel * d_tr * pa_pr + \
        -1.24382300e-5 * vel * vel * vel * d_tr * pa_pr + \
        -7.38584400e-9 * ta * vel * vel * vel * d_tr * pa_pr + \
        2.20609296e-7 * vel * vel * vel * vel * d_tr * pa_pr + \
        -7.32469180e-4 * d_tr * d_tr * pa_pr + \
        -1.87381964e-5 * ta * d_tr * d_tr * pa_pr + \
        4.80925239e-6 * ta * ta * d_tr * d_tr * pa_pr + \
        -8.75492040e-8 * ta * ta * ta * d_tr * d_tr * pa_pr + \
        2.77862930e-5 * vel * d_tr * d_tr * pa_pr + \
        -5.06004592e-6 * ta * vel * d_tr * d_tr * pa_pr + \
        1.14325367e-7 * ta * ta * vel * d_tr * d_tr * pa_pr + \
        2.53016723e-6 * vel * vel * d_tr * d_tr * pa_pr + \
        -1.72857035e-8 * ta * vel * vel * d_tr * d_tr * pa_pr + \
        -3.95079398e-8 * vel * vel * vel * d_tr * d_tr * pa_pr + \
        -3.59413173e-7 * d_tr * d_tr * d_tr * pa_pr + \
        7.04388046e-7 * ta * d_tr * d_tr * d_tr * pa_pr + \
        -1.89309167e-8 * ta * ta * d_tr * d_tr * d_tr * pa_pr + \
        -4.79768731e-7 * vel * d_tr * d_tr * d_tr * pa_pr + \
        7.96079978e-9 * ta * vel * d_tr * d_tr * d_tr * pa_pr + \
        1.62897058e-9 * vel * vel * d_tr * d_tr * d_tr * pa_pr + \
        3.94367674e-8 * d_tr * d_tr * d_tr * d_tr * pa_pr + \
        -1.18566247e-9 * ta * d_tr * d_tr * d_tr * d_tr * pa_pr + \
        3.34678041e-10 * vel * d_tr * d_tr * d_tr * d_tr * pa_pr + \
        -1.15606447e-10 * d_tr * d_tr * d_tr * d_tr * d_tr * pa_pr + \
        -2.80626406 * pa_pr * pa_pr + \
        0.548712484 * ta * pa_pr * pa_pr + \
        -0.00399428410 * ta * ta * pa_pr * pa_pr + \
        -9.54009191e-4 * ta * ta * ta * pa_pr * pa_pr + \
        1.93090978e-5 * ta * ta * ta * ta * pa_pr * pa_pr + \
        -0.308806365 * vel * pa_pr * pa_pr + \
        0.0116952364 * ta * vel * pa_pr * pa_pr + \
        4.95271903e-4 * ta * ta * vel * pa_pr * pa_pr + \
        -1.90710882e-5 * ta * ta * ta * vel * pa_pr * pa_pr + \
        0.00210787756 * vel * vel * pa_pr * pa_pr + \
        -6.98445738e-4 * ta * vel * vel * pa_pr * pa_pr + \
        2.30109073e-5 * ta * ta * vel * vel * pa_pr * pa_pr + \
        4.17856590e-4 * vel * vel * vel * pa_pr * pa_pr + \
        -1.27043871e-5 * ta * vel * vel * vel * pa_pr * pa_pr + \
        -3.04620472e-6 * vel * vel * vel * vel * pa_pr * pa_pr + \
        0.0514507424 * d_tr * pa_pr * pa_pr + \
        -0.00432510997 * ta * d_tr * pa_pr * pa_pr + \
        8.99281156e-5 * ta * ta * d_tr * pa_pr * pa_pr + \
        -7.14663943e-7 * ta * ta * ta * d_tr * pa_pr * pa_pr + \
        -2.66016305e-4 * vel * d_tr * pa_pr * pa_pr + \
        2.63789586e-4 * ta * vel * d_tr * pa_pr * pa_pr + \
        -7.01199003e-6 * ta * ta * vel * d_tr * pa_pr * pa_pr + \
        -1.06823306e-4 * vel * vel * d_tr * pa_pr * pa_pr + \
        3.61341136e-6 * ta * vel * vel * d_tr * pa_pr * pa_pr + \
        2.29748967e-7 * vel * vel * vel * d_tr * pa_pr * pa_pr + \
        3.04788893e-4 * d_tr * d_tr * pa_pr * pa_pr + \
        -6.42070836e-5 * ta * d_tr * d_tr * pa_pr * pa_pr + \
        1.16257971e-6 * ta * ta * d_tr * d_tr * pa_pr * pa_pr + \
        7.68023384e-6 * vel * d_tr * d_tr * pa_pr * pa_pr + \
        -5.47446896e-7 * ta * vel * d_tr * d_tr * pa_pr * pa_pr + \
        -3.59937910e-8 * vel * vel * d_tr * d_tr * pa_pr * pa_pr + \
        -4.36497725e-6 * d_tr * d_tr * d_tr * pa_pr * pa_pr + \
        1.68737969e-7 * ta * d_tr * d_tr * d_tr * pa_pr * pa_pr + \
        2.67489271e-8 * vel * d_tr * d_tr * d_tr * pa_pr * pa_pr + \
        3.23926897e-9 * d_tr * d_tr * d_tr * d_tr * pa_pr * pa_pr + \
        -0.0353874123 * pa_pr * pa_pr * pa_pr + \
        -0.221201190 * ta * pa_pr * pa_pr * pa_pr + \
        0.0155126038 * ta * ta * pa_pr * pa_pr * pa_pr + \
        -2.63917279e-4 * ta * ta * ta * pa_pr * pa_pr * pa_pr + \
        0.0453433455 * vel * pa_pr * pa_pr * pa_pr + \
        -0.00432943862 * ta * vel * pa_pr * pa_pr * pa_pr + \
        1.45389826e-4 * ta * ta * vel * pa_pr * pa_pr * pa_pr + \
        2.17508610e-4 * vel * vel * pa_pr * pa_pr * pa_pr + \
        -6.66724702e-5 * ta * vel * vel * pa_pr * pa_pr * pa_pr + \
        3.33217140e-5 * vel * vel * vel * pa_pr * pa_pr * pa_pr + \
        -0.00226921615 * d_tr * pa_pr * pa_pr * pa_pr + \
        3.80261982e-4 * ta * d_tr * pa_pr * pa_pr * pa_pr + \
        -5.45314314e-9 * ta * ta * d_tr * pa_pr * pa_pr * pa_pr + \
        -7.96355448e-4 * vel * d_tr * pa_pr * pa_pr * pa_pr + \
        2.53458034e-5 * ta * vel * d_tr * pa_pr * pa_pr * pa_pr + \
        -6.31223658e-6 * vel * vel * d_tr * pa_pr * pa_pr * pa_pr + \
        3.02122035e-4 * d_tr * d_tr * pa_pr * pa_pr * pa_pr + \
        -4.77403547e-6 * ta * d_tr * d_tr * pa_pr * pa_pr * pa_pr + \
        1.73825715e-6 * vel * d_tr * d_tr * pa_pr * pa_pr * pa_pr + \
        -4.09087898e-7 * d_tr * d_tr * d_tr * pa_pr * pa_pr * pa_pr + \
        0.614155345 * pa_pr * pa_pr * pa_pr * pa_pr + \
        -0.0616755931 * ta * pa_pr * pa_pr * pa_pr * pa_pr + \
        0.00133374846 * ta * ta * pa_pr * pa_pr * pa_pr * pa_pr + \
        0.00355375387 * vel * pa_pr * pa_pr * pa_pr * pa_pr + \
        -5.13027851e-4 * ta * vel * pa_pr * pa_pr * pa_pr * pa_pr + \
        1.02449757e-4 * vel * vel * pa_pr * pa_pr * pa_pr * pa_pr + \
        -0.00148526421 * d_tr * pa_pr * pa_pr * pa_pr * pa_pr + \
        -4.11469183e-5 * ta * d_tr * pa_pr * pa_pr * pa_pr * pa_pr + \
        -6.80434415e-6 * vel * d_tr * pa_pr * pa_pr * pa_pr * pa_pr + \
        -9.77675906e-6 * d_tr * d_tr * pa_pr * pa_pr * pa_pr * pa_pr + \
        0.0882773108 * pa_pr * pa_pr * pa_pr * pa_pr * pa_pr + \
        -0.00301859306 * ta * pa_pr * pa_pr * pa_pr * pa_pr * pa_pr + \
        0.00104452989 * vel * pa_pr * pa_pr * pa_pr * pa_pr * pa_pr + \
        2.47090539e-4 * d_tr * pa_pr * pa_pr * pa_pr * pa_pr * pa_pr + \
        0.00148348065 * pa_pr * pa_pr * pa_pr * pa_pr * pa_pr * pa_pr

    return utci_approx


def saturated_vapor_pressure_hpa(db_temp):
    """Calculate saturated vapor pressure (hPa) at temperature (C).

    This equation of saturation vapor pressure is specific to the UTCI model.
    """
    g = (-2836.5744, -6028.076559, 19.54263612, -0.02737830188, 0.000016261698,
         7.0229056e-10, -1.8680009e-13)
    tk = db_temp + 273.15  # air temp in K
    es = 2.7150305 * math.log(tk)
    for i, x in enumerate(g):
        es = es + (x * (tk**(i - 2)))
    es = math.exp(es) * 0.01
    return es


def calc_missing_utci_input(target_utci, utci_inputs,
                            low_bound=0., up_bound=100., tolerance=0.001):
    """Return the value of a missing_utci_input given a target_utci and the 3 other inputs.

    This is particularly useful when trying to draw comfort polygons on charts
    using the UTCI model.

    Args:
        target_utci: The target UTCI temperature that you are trying to produce
            from the inputs to the UTCI model.
        utci_inputs: A dictionary of 4 UTCI inputs with the following keys:
            'ta', 'tr', 'vel', 'rh'.  Each key should correspond to a value
            that represents that UTCI input but one of these inputs should
            have a value of None.
            The input corresponding to None will be solved for by this function.
            One can also input None for both 'ta' and 'tr' to solve for the operative
            temperature that meets the target_utci. In this case, both 'ta' and 'tr'
            in the output dictionary will be the same.
            Example (solving for relative humidity):

         .. code-block:: python

            {'ta': 20, 'tr': 20, 'vel': 0.05, 'rh': None}

        low_bound: The lowest possible value of the missing input you are tying to
            find. Putting in a good value here will help the model converge to a
            solution faster.
        up_bound: The highest possible value of the missing input you are tying to
            find. Putting in a good value here will help the model converge to a
            solution faster.
        tolerance: The acceptable error in the target_utci. The default is set to 0.001

    Returns:
        complete_utci_inputs -- The utci_inputs dictionary but with values for
        all inputs. The missing input to the UTCI model will be filled by the value
        that returns the target_utci.
    """
    assert len(utci_inputs.keys()) == 4, \
        'utci_inputs must have 4 keys. Got {}.'.format(len(utci_inputs.keys()))

    # Determine the function that should be used given the missing input.
    if utci_inputs['ta'] is None and utci_inputs['tr'] is None:
        def fn(x):
            return universal_thermal_climate_index(
                x, x, utci_inputs['vel'], utci_inputs['rh']
                ) - target_utci
        missing_key = ('ta', 'tr')
    elif utci_inputs['ta'] is None:
        def fn(x):
            return universal_thermal_climate_index(
                x, utci_inputs['tr'], utci_inputs['vel'], utci_inputs['rh']
                ) - target_utci
        missing_key = 'ta'
    elif utci_inputs['tr'] is None:
        def fn(x):
            return universal_thermal_climate_index(
                utci_inputs['ta'], x, utci_inputs['vel'], utci_inputs['rh']
                ) - target_utci
        missing_key = 'tr'
    elif utci_inputs['vel'] is None:
        def fn(x):
            return target_utci - universal_thermal_climate_index(
                utci_inputs['ta'], utci_inputs['tr'], x, utci_inputs['rh'])
        missing_key = 'vel'
    else:
        def fn(x):
            return universal_thermal_climate_index(
                utci_inputs['ta'], utci_inputs['tr'], utci_inputs['vel'], x
                ) - target_utci
        missing_key = 'rh'

    # Solve for the missing input using the function.
    missing_val = secant(low_bound, up_bound, fn, tolerance)
    if missing_val is None:
        missing_val = bisect(low_bound, up_bound, fn, tolerance, 0)

    # complete the input dictionary
    if isinstance(missing_key, str):
        utci_inputs[missing_key] = missing_val
    else:
        for key in missing_key:
            utci_inputs[key] = missing_val
    return utci_inputs
